from ngsolve import *
from netgen.geom2d import SplineGeometry
from numpy import pi
ngsglobals.msg_level=1

def topBottom():
  geometry = SplineGeometry()
  
  # point coordinates ...
  pnts = [ (0,0), (1,0),  (1,0.5), (1,1), (0,1), (0, 0.5)]
  pnums = [geometry.AppendPoint(*p) for p in pnts]
  # start-point, end-point, boundary-condition, domain on left side, domain on right side:
  lines = [ (pnums[0],pnums[1],"gammaD",2,0), (pnums[1],pnums[2],"gammaD",2,0), 
       (pnums[2],pnums[3],"gammaS",1,0), (pnums[3],pnums[4],"gammaS",1,0),
       (pnums[4],pnums[5],"gammaS",1,0),
       (pnums[5],pnums[0],"gammaD",2,0), (pnums[5],pnums[2],"gammaSD",1,2)]
  for p1,p2,bc,left,right in lines:
      geometry.Append( ["line", p1, p2], bc=bc, leftdomain=left, rightdomain=right)
  return geometry

def GetInterfaceIndicator(mesh):
  dom_indicator = GridFunction(L2(mesh,order=0))
  dom_indicator.Set(CoefficientFunction([1,2]))
  
  ffes = FacetFESpace(mesh,order=0)
  dom_ind_average = GridFunction(ffes)
  mf = BilinearForm(ffes,symmetric=True)
  mf += SymbolicBFI(ffes.TrialFunction()*ffes.TestFunction(),element_boundary=True)
  mf.Assemble()
  ff = LinearForm(ffes)
  ff += SymbolicLFI(dom_indicator*ffes.TestFunction(),element_boundary=True)
  ff.Assemble()
  dom_ind_average.vec.data = mf.mat.Inverse() * ff.vec

  interface_indicator = BitArray(ffes.ndof)
  for i in range(len(dom_ind_average.vec)):
    if (dom_ind_average.vec[i] > 1.75) or (dom_ind_average.vec[i] < 1.25):
      interface_indicator[i] = False
    else:
      interface_indicator[i] = True
  print(sum(interface_indicator))
  return interface_indicator

def GetFaceIndicator(mesh):
  dom_indicator = GridFunction(L2(mesh,order=0))
  dom_indicator.Set(CoefficientFunction([1,2]))
  
  ffes = FacetFESpace(mesh,order=0)
  dom_ind_average = GridFunction(ffes)
  mf = BilinearForm(ffes,symmetric=True)
  mf += SymbolicBFI(ffes.TrialFunction()*ffes.TestFunction(),element_boundary=True)
  mf.Assemble()
  ff = LinearForm(ffes)
  ff += SymbolicLFI(dom_indicator*ffes.TestFunction(),element_boundary=True)
  ff.Assemble()
  dom_ind_average.vec.data = mf.mat.Inverse() * ff.vec

  stokes_indicator = BitArray(ffes.ndof)
  darcy_indicator = BitArray(ffes.ndof)
  for i in range(len(dom_ind_average.vec)):
    if (dom_ind_average.vec[i] > 1.75): # exclude darcy part
      stokes_indicator[i] = False
    else:
      stokes_indicator[i] = True
    if (dom_ind_average.vec[i] < 1.25): # exclude darcy part
      darcy_indicator[i] = False
    else:
      darcy_indicator[i] = True
  #print("facets marked as stokes facets:", stokes_indicator)
  #print("facets marked as darcy facets:", darcy_indicator)
  return stokes_indicator, darcy_indicator


stokesId = [1,0]
darcyId = [0,1]

nu = 1.
kinv = 1.
gamma = (1.+4.*pi*pi)/2.
order = 4
hmax = 1.0/8
dcFlag = False

n = specialcf.normal(2)
h = specialcf.mesh_size
penalty = 10*order*(order+1)/h
ndofMax = 10

orderp = order-1

mesh = Mesh(topBottom().GenerateMesh (maxh=hmax))

# exact solution
u1 = sin(pi*x)*exp(y/2.)
u2 = cos(pi*x)*exp(y/2.)
uexS = CoefficientFunction((-1./2./pi/pi*u1,1./pi*u2))
uexD = CoefficientFunction((-2.*u1,1./pi*u2))
pexS = CoefficientFunction(-1./pi*u2)
pexD = CoefficientFunction(-2./pi*u2)

cst1 = 1.+nu*(-.5+1./8./pi/pi)
cst2 = -.5/pi+nu*(pi-.25/pi)
cst3 = 0.5/pi-2.*pi
stokesSource = CoefficientFunction((cst1*u1,cst2*u2))
darcySource = CoefficientFunction(cst3*u2)



dom1 = GridFunction(L2(mesh,order=0))
dom2 = GridFunction(L2(mesh,order=0))

V1 = HDiv(mesh, order=order, dirichlet="gammaS|gammaD")
# FIXME define trace space only on stokes domain
V2 = FESpace ("vectorfacet", mesh, order=order, dirichlet="gammaS",definedon = [1], flags
    ={"highest_order_dc": dcFlag})
V3 = L2(mesh, order=orderp)
V4 = FESpace("number",mesh)

V = FESpace([V1,V2, V3,V4], flags={"dgjumps": True})
u,uhat, p, pbar = V.TrialFunction()
v,vhat, q, qbar = V.TestFunction()
crossTerm = -div(u)*q-div(v)*p + p*qbar + q*pbar

du = grad(u)
dv = grad(v)
gradu = CoefficientFunction ( (du,), dims=(2,2) )
gradv = CoefficientFunction ( (dv,), dims=(2,2) )
epsu = 0.5*(gradu+gradu.trans)
epsv = 0.5*(gradv+gradv.trans)

def tang(vec):
    return vec - (vec*n)*n
interface_indicator = GetInterfaceIndicator(mesh)


a = BilinearForm(V)
# dg stokes
stokesCell = 2*nu*InnerProduct(epsu,epsv)
stokesFacet1 = 2*nu*InnerProduct ( epsu*n,  tang(vhat-v) )
stokesFacet2 = 2*nu*InnerProduct ( epsv*n,  tang(uhat-u) )
stokesFacet3 = penalty*nu*InnerProduct (tang(vhat-v),tang(uhat-u) )

# Hdiv darcy
darcyCell = kinv*u*v

# interface (only stokes part is needed)
interfaceTerm = gamma*sqrt(kinv)*(tang(uhat)*tang(vhat))


## DEBUGGING
# Stokes is correct
a += SymbolicBFI ( stokesCell+dom2*darcyCell + crossTerm, definedon=[1])
a += SymbolicBFI ( darcyCell + crossTerm, definedon = [2])
a += SymbolicBFI ( (stokesFacet1+stokesFacet2+stokesFacet3), element_boundary=True, definedon=[1])
# FIXME change to element_boundary 
interfacebfi = SymbolicBFI (interfaceTerm, VOL, skeleton=True) 
interfacebfi.SetDefinedOnElements(interface_indicator)
a += interfacebfi


a.Assemble()

c = Preconditioner(type="direct", bf=a, flags = { "inverse" : "pardiso" } )

f = LinearForm(V)
f += SymbolicLFI(dom1*stokesSource*v-dom2*darcySource*q)
f.Assemble()

u = GridFunction(V)
uS = GridFunction(V)
uD = GridFunction(V)
bvp =  BVP(bf=a,lf=f,gf=u,pre=c, maxsteps=1000, prec=1e-11)
vtk = VTKOutput(ma=mesh,coefs=[u.components[0],
  u.components[2]],names=["velocity","pressure"],filename="hdg",subdivision=3)
def solve():
    dom1.Update()
    dom1.Set(CoefficientFunction(stokesId))
    dom2.Update()
    dom2.Set(CoefficientFunction(darcyId))
    V.Update()
    
    interface_indicator = GetInterfaceIndicator(mesh)
    #ss, interface_indicator = GetFaceIndicator(mesh)
    interfacebfi.SetDefinedOnElements(interface_indicator)
    
    a.Assemble()
    f.Assemble()
    u.Update()
    uS.Update()
    uD.Update()
    #uS.components[0].Set(uexS,definedon=mesh.Boundaries("gammaS"))
    #uS.components[1].Set(uexS,definedon=mesh.Boundaries("gammaS"))
    #uD.components[0].Set(uexD,definedon=mesh.Boundaries("gammaD"))
    #u.vec.data = uS.vec+uD.vec
    u.components[0].Set(dom1*uexS+dom2*uexD)
    u.components[1].Set(uexS,definedon=mesh.Boundaries("gammaS"))

    #res = f.vec.CreateVector()
    #res.data = f.vec - a.mat*u.vec
    #u.vec.data += a.mat.Inverse(V.FreeDofs()) * res
    bvp.Do()
    uS.components[0].Set(dom1*uexS+dom2*uexD)
    uD.vec.data = uS.vec-u.vec
    #Draw(uD.components[0])
    errS = dom1*(u.components[0]-uexS)*(u.components[0]-uexS)
    errD = dom2*(u.components[0]-uexD)*(u.components[0]-uexD)
    erru1 = Integrate(errS,mesh,order = 2*order)
    erru2 = Integrate(errD,mesh,order = 2*order)
    return sqrt(erru1), sqrt(erru2) 
l= []
erru1, erru2 = solve()
print(erru1, erru2)
l.append([erru1, erru2])
with TaskManager():
    while V.ndof < ndofMax:  
        mesh.Refine()
        erru1, erru2 = solve()
        l.append([erru1, erru2])
        print(erru1, erru2)
        Redraw()


for i in range(len(l)-1):
    rate = -log(l[i+1][0]/l[i][0])/log(2)
    print("%2.2e %2.2f "%(l[i][0], rate))
